//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
using System;

using Do.Addins;

namespace JIRA
{
	public interface IJIRAConfiguration
	{
		string BaseUrl
		{
			get;
		}
		
		string Username
		{
			get;
		}
		
		string Password 
		{
			get;
		}
		
		string[] Projects 
		{ 
			get;
		}
		
		/// <value>
		/// To be valid, all fields must be filled out
		/// </value>
		bool IsValid();
	}	
	
	/// <summary>
	/// Configuration that uses gconf to do the magic
	/// </summary>
	public class JIRAConfiguration : IJIRAConfiguration
	{
		/// <summary>
		/// The gnome conf key that contains the base url for our JIRA installation.
		/// ie. http://issues.apache.org/jira
		/// </summary>
		private const string _sGCONF_KEY= "/apps/gnome-do/plugins/JIRA";
	
		private GConf.Client _gconf;		
		
		public JIRAConfiguration()
		{
			_gconf= new GConf.Client();
		}
		
		private void SetConfValue( string key, string val )
		{
			_gconf.Set( _sGCONF_KEY+"/"+key, val );
		}
		
		private string GetConfValue( string key, string fallback )
		{
			try
			{
				return _gconf.Get( _sGCONF_KEY+"/"+key ) as String;
			}
			catch( GConf.NoSuchKeyException ) 
			{
				return fallback;
			}
		}
		
		public string BaseUrl
		{
			get {  return GetConfValue( "baseUrl", "" );	}
			set { SetConfValue( "baseUrl", value ); }
		}
		
		public string Username
		{
			get {  return GetConfValue( "username", "" );	}
			set { SetConfValue( "username", value ); }
		}
		
		public string Password 
		{
			get {  return GetConfValue( "password", "" );	}
			set { SetConfValue( "password", value ); }
		}
		public string[] Projects 
		{ 
			get 
			{
				string projStr= GetConfValue( "projects", "" );
				
				return projStr.Length>0 ? projStr.Split( ',' ) : new string[ 0 ];
			}
			set { SetConfValue( "projects", string.Join( ",", value ) ); }
		}
		
		/// <value>
		/// To be valid, all fields must be filled out
		/// </value>
		public bool IsValid()
		{
			return BaseUrl.Length>0
				&& Username.Length>0
				&& Password.Length>0
				&& Projects.Length>0; 
		}
	}
}
