//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

using System;
using Do.Universe;
using System.Collections.Generic;

namespace SqueezeCenter
{
	
	public abstract class PlayerAction : AbstractAction
	{
		string name;
		string description;
		string icon;
		PlayerStatus requiredPlayerStatus;
		
		public PlayerAction (string name, string description, string icon, PlayerStatus requiredPlayerStatus)
		{
			this.name = name;
			this.description = description;
			this.icon = icon;
			this.requiredPlayerStatus = requiredPlayerStatus;
		}
		
		public override string Name {
			get { return name; }
		}

		public override string Description {
			get { return description; }
		}

		public override string Icon {
			get { return icon; }
		}

		public override Type[] SupportedItemTypes {
			get {
				return new Type[] {
					typeof (Player),
				};
			}
		}
		
		public override bool SupportsItem (IItem item)
		{
			return (((item as Player).PoweredOn ? 1 : 2) & (int)requiredPlayerStatus) > 0;			
		}
		
		public abstract string GetCommand (Player player, IItem modifierItem);
		
		public override IItem[] Perform (IItem[] items, IItem[] modifierItems)
		{
			string command = GetCommand ((Player)items[0], modifierItems.Length > 0 ? modifierItems[0] : null);
			Server.Instance.ExecuteCommand (command);
		
			return null;
		}
	}
	
	
	
	public class TurnOn : PlayerAction 
	{
		
		public TurnOn () : base("Turn on", "Turn on the player", "sunny", PlayerStatus.Off) {}
		
		public override string GetCommand (Player player, IItem modifierItem)
		{
			return string.Format ("{0} power 1", player.Id);
		}
	}
	
	public class TurnOff : PlayerAction 
	{
		
		public TurnOff () : base("Turn off", "Turn off the player", "gnome-shutdown", PlayerStatus.On) {}
		
		public override string GetCommand (Player player, IItem modifierItem)
		{
			return string.Format ("{0} power 0", player.Id);
		}
		
	}
	
	public class Pause : PlayerAction 
	{
		
		public Pause () : base ("Pause", "Toggle pause on the player", "gtk-media-pause", PlayerStatus.On) {}
		
		public override string GetCommand (Player player, IItem modifierItem)
		{
			return string.Format ("{0} pause", player.Id);
		}
		
	}
	
	public class Sync : PlayerAction 
	{
		
		public Sync () : base("Sync with", "Synchronize this player with another", 
		//                      "sync.png@" + typeof (Sync).Assembly.FullName,
		                      "sync.png@SqueezeCenterPlugin",
		                      PlayerStatus.Any) {}
		
		public override Type[] SupportedModifierItemTypes {
			get { 
				return new Type[] {
					typeof (Player),
				};
			}
		}

		public override bool SupportsModifierItemForItems (IItem[] items, IItem modItem)
		{
			// all players except the current one
			return Array.IndexOf (items, modItem) < 0;			
		}
		
		public override bool ModifierItemsOptional
		{
			get {
				return false;
			}
		}
		
		public override string GetCommand (Player player, IItem modifierItem)
		{
			return string.Format ("{1} sync {0}", player.Id, (modifierItem as Player).Id);
		}
	}
	
	public class Unsync : PlayerAction		
	{
		
		public Unsync () : base("Unsync", "Stop synchronizing this player with another", 
		                        //"unsync.png@" + typeof (Unsync).Assembly.FullName, 
		                        "unsync.png@SqueezeCenterPlugin",
		                        PlayerStatus.Any) {}
		
		public override string GetCommand (Player player, IItem modifierItem)
		{
			return string.Format ("{0} sync -", player.Id);
		}
		
		public override bool SupportsModifierItemForItems (IItem[] items, IItem modItem)
		{			
			return false;
		}
		
		public override bool SupportsItem (IItem item)
		{
			return base.SupportsItem (item) && (item as Player).IsSynced;
		}
 
		
		public override bool ModifierItemsOptional
		{
			get {
				return true;
			}
		}		
	}
	
	public class Prev : PlayerAction 
	{
		
		public Prev () : base("Previous", "Previous track", "previous", PlayerStatus.On) {}
		
		public override string GetCommand (Player player, IItem modifierItem)
		{
			return string.Format ("{0} playlist index -1", player.Id);
		}
		
	}
	
	public class Next : PlayerAction 
	{
		
		public Next () : base("Next", "Next track", "next", PlayerStatus.On) {}
		
		public override string GetCommand (Player player, IItem modifierItem)
		{
			return string.Format ("{0} playlist index +1", player.Id);
		}
		
	}	
	
	public enum PlayerStatus : int 
	{
		On = 1,
		Off = 2,
		Any = On | Off
	}
	
}
